@@ -23,9 +23,9 @@
 #endif
 
 #include <map>
-#include "behaviortree_cpp_v3/xml_parsing.h"
+#include "behaviortree_cpp/xml_parsing.h"
 #include "tinyxml2/tinyxml2.h"
-#include "filesystem/path.h"
+#include <filesystem>
 
 #ifdef USING_ROS
 #include <ros/package.h>
@@ -35,8 +35,9 @@
 #include <ament_index_cpp/get_package_share_directory.hpp>
 #endif
 
-#include "behaviortree_cpp_v3/blackboard.h"
-#include "behaviortree_cpp_v3/utils/demangle_util.h"
+#include "behaviortree_cpp/blackboard.h"
+#include "behaviortree_cpp/tree_node.h"
+#include "behaviortree_cpp/utils/demangle_util.h"
 
 namespace BT
 {
@@ -46,15 +47,20 @@ auto StrEqual = [](const char* str1, const char* str2) -> bool {
   return strcmp(str1, str2) == 0;
 };
 
-struct XMLParser::Pimpl
+struct XMLParser::PImpl
 {
   TreeNode::Ptr createNodeFromXML(const XMLElement* element,
                                   const Blackboard::Ptr& blackboard,
-                                  const TreeNode::Ptr& node_parent);
+                                  const TreeNode::Ptr& node_parent,
+                                  const std::string &prefix_path,
+                                  Tree& output_tree);
 
-  void recursivelyCreateTree(const std::string& tree_ID, Tree& output_tree,
-                             Blackboard::Ptr blackboard,
-                             const TreeNode::Ptr& root_parent);
+  void recursivelyCreateSubtree(const std::string& tree_ID,
+                                const std::string &tree_name,
+                                const std::string &prefix_path,
+                                Tree& output_tree,
+                                Blackboard::Ptr blackboard,
+                                const TreeNode::Ptr& root_node);
 
   void getPortsRecursively(const XMLElement* element,
                            std::vector<std::string>& output_ports);
@@ -66,18 +72,18 @@ struct XMLParser::Pimpl
 
   const BehaviorTreeFactory& factory;
 
-  filesystem::path current_path;
+  std::filesystem::path current_path;
 
   int suffix_count;
-
-  explicit Pimpl(const BehaviorTreeFactory& fact) :
-    factory(fact), current_path(filesystem::path::getcwd()), suffix_count(0)
+  
+  explicit PImpl(const BehaviorTreeFactory& fact) :
+    factory(fact), current_path(std::filesystem::current_path()), suffix_count(0)
   {}
 
   void clear()
   {
     suffix_count = 0;
-    current_path = filesystem::path::getcwd();
+    current_path = std::filesystem::current_path();
     opened_documents.clear();
     tree_roots.clear();
   }
@@ -87,23 +93,31 @@ struct XMLParser::Pimpl
 #pragma GCC diagnostic pop
 #endif
 
-XMLParser::XMLParser(const BehaviorTreeFactory& factory) : _p(new Pimpl(factory))
+XMLParser::XMLParser(const BehaviorTreeFactory& factory) : _p(new PImpl(factory))
 {}
 
-XMLParser::~XMLParser()
+XMLParser::XMLParser(XMLParser &&other) noexcept
 {
-  delete _p;
+  this->_p = std::move(other._p);
 }
 
-void XMLParser::loadFromFile(const std::string& filename, bool add_includes)
+XMLParser &XMLParser::operator=(XMLParser &&other) noexcept
+{
+  this->_p = std::move(other._p);
+  return *this;
+}
+
+XMLParser::~XMLParser()
+{}
+
+void XMLParser::loadFromFile(const std::filesystem::path& filepath, bool add_includes)
 {
   _p->opened_documents.emplace_back(new tinyxml2::XMLDocument());
 
   tinyxml2::XMLDocument* doc = _p->opened_documents.back().get();
-  doc->LoadFile(filename.c_str());
+  doc->LoadFile(filepath.string().c_str());
 
-  filesystem::path file_path(filename);
-  _p->current_path = file_path.parent_path().make_absolute();
+  _p->current_path = std::filesystem::absolute(filepath.parent_path());
 
   _p->loadDocImpl(doc, add_includes);
 }
@@ -128,36 +142,41 @@ std::vector<std::string> XMLParser::registeredBehaviorTrees() const
   return out;
 }
 
-void XMLParser::Pimpl::loadDocImpl(tinyxml2::XMLDocument* doc, bool add_includes)
+void XMLParser::PImpl::loadDocImpl(tinyxml2::XMLDocument* doc, bool add_includes)
 {
   if (doc->Error())
   {
     char buffer[200];
-    sprintf(buffer, "Error parsing the XML: %s", doc->ErrorName());
+    sprintf(buffer, "Error parsing the XML: %s", doc->ErrorStr());
     throw RuntimeError(buffer);
   }
 
   const XMLElement* xml_root = doc->RootElement();
 
+  auto format = xml_root->Attribute("BTCPP_format");
+  if(!format)
+  {
+    std::cout << "Warnings: The first tag of the XML (<root>) should contain the attribute [BTCPP_format=\"4\"]\n"
+              << "Please check if your XML is compatible with version 4.x of BT.CPP" << std::endl;
+  }
+
   // recursively include other files
-  for (auto include_node = xml_root->FirstChildElement("include");
-       include_node != nullptr; include_node = include_node->NextSiblingElement("includ"
-                                                                                "e"))
+  for (auto incl_node = xml_root->FirstChildElement("include"); incl_node != nullptr;
+       incl_node = incl_node->NextSiblingElement("include"))
   {
     if (!add_includes)
     {
       break;
     }
 
-    filesystem::path file_path(include_node->Attribute("path"));
-    const char* ros_pkg_relative_path = include_node->Attribute("ros_pkg");
+    std::filesystem::path file_path(incl_node->Attribute("path"));
+    const char* ros_pkg_relative_path = incl_node->Attribute("ros_pkg");
 
     if (ros_pkg_relative_path)
     {
       if (file_path.is_absolute())
       {
-        std::cout << "WARNING: <include path=\"...\"> contains an absolute "
-                     "path.\n"
+        std::cout << "WARNING: <include path=\"...\"> contains an absolute path.\n"
                   << "Attribute [ros_pkg] will be ignored." << std::endl;
       }
       else
@@ -165,19 +184,15 @@ void XMLParser::Pimpl::loadDocImpl(tinyxml2::XMLDocument* doc, bool add_includes
         std::string ros_pkg_path;
 #ifdef USING_ROS
         ros_pkg_path = ros::package::getPath(ros_pkg_relative_path);
-        file_path = filesystem::path(ros_pkg_path) / file_path;
 #elif defined USING_ROS2
         ros_pkg_path =
             ament_index_cpp::get_package_share_directory(ros_pkg_relative_path);
-        file_path = filesystem::path(ros_pkg_path) / file_path;
 #else
-        throw RuntimeError("Using attribute [ros_pkg] in <include>, but this "
-                           "library was "
-                           "compiled "
-                           "without ROS support. Recompile the BehaviorTree.CPP "
-                           "using "
-                           "catkin");
+        throw RuntimeError("Using attribute [ros_pkg] in <include>, but this library was "
+                           "compiled without ROS support. Recompile the BehaviorTree.CPP "
+                           "using catkin");
 #endif
+        file_path = std::filesystem::path(ros_pkg_path) / file_path;
       }
     }
 
@@ -190,10 +205,10 @@ void XMLParser::Pimpl::loadDocImpl(tinyxml2::XMLDocument* doc, bool add_includes
     tinyxml2::XMLDocument* next_doc = opened_documents.back().get();
 
     // change current path to the included file for handling additional relative paths
-    const filesystem::path previous_path = current_path;
-    current_path = file_path.parent_path().make_absolute();
+    const auto previous_path = current_path;
+    current_path = std::filesystem::absolute(file_path.parent_path());
 
-    next_doc->LoadFile(file_path.str().c_str());
+    next_doc->LoadFile(file_path.string().c_str());
     loadDocImpl(next_doc, add_includes);
 
     // reset current path to the previous value
@@ -228,7 +243,7 @@ void XMLParser::Pimpl::loadDocImpl(tinyxml2::XMLDocument* doc, bool add_includes
       tree_name = "BehaviorTree_" + std::to_string(suffix_count++);
     }
 
-    tree_roots.insert({tree_name, bt_node});
+    tree_roots[tree_name] = bt_node;
   }
 }
 
@@ -294,8 +309,7 @@ void VerifyXML(const std::string& xml_text,
         if (!ID)
         {
           ThrowError(node->GetLineNum(), "Error at line %d: -> The attribute "
-                                         "[ID] is "
-                                         "mandatory");
+                                         "[ID] is mandatory");
         }
       }
     }
@@ -437,10 +451,9 @@ void VerifyXML(const std::string& xml_text,
 }
 
 Tree XMLParser::instantiateTree(const Blackboard::Ptr& root_blackboard,
-                                std::string main_tree_to_execute)
+                                std::string main_tree_ID)
 {
   Tree output_tree;
-  std::string main_tree_ID = main_tree_to_execute;
 
   // use the main_tree_to_execute argument if it was provided by the user
   // or the one in the FIRST document opened
@@ -469,104 +482,150 @@ Tree XMLParser::instantiateTree(const Blackboard::Ptr& root_blackboard,
     throw RuntimeError("XMLParser::instantiateTree needs a non-empty "
                        "root_blackboard");
   }
-  // first blackboard
-  output_tree.blackboard_stack.push_back(root_blackboard);
 
-  _p->recursivelyCreateTree(main_tree_ID, output_tree, root_blackboard, TreeNode::Ptr());
+  _p->recursivelyCreateSubtree(main_tree_ID, {}, {},
+                               output_tree, root_blackboard, TreeNode::Ptr());
   output_tree.initialize();
   return output_tree;
 }
 
 void XMLParser::clearInternalState()
 {
-    _p->clear();
+  _p->clear();
 }
 
-TreeNode::Ptr XMLParser::Pimpl::createNodeFromXML(const XMLElement* element,
+TreeNode::Ptr XMLParser::PImpl::createNodeFromXML(const XMLElement* element,
                                                   const Blackboard::Ptr& blackboard,
-                                                  const TreeNode::Ptr& node_parent)
+                                                  const TreeNode::Ptr& node_parent,
+                                                  const std::string& prefix_path,
+                                                  Tree& output_tree)
 {
-  const std::string element_name = element->Name();
-  std::string ID;
-  std::string instance_name;
+  auto element_name = element->Name();
+  auto element_ID = element->Attribute("ID");
 
-  // Actions and Decorators have their own ID
-  if (element_name == "Action" || element_name == "Decorator" ||
-      element_name == "Condition" || element_name == "Control")
+  auto node_type = convertFromString<NodeType>(element_name);
+  // name used by the factory
+  std::string type_ID;
+
+  if (node_type == NodeType::UNDEFINED)
   {
-    ID = element->Attribute("ID");
+    // This is the case of nodes like <MyCustomAction>
+    // check if the factory has this name
+    if (factory.builders().count(element_name) == 0)
+    {
+      throw RuntimeError(element_name, " is not a registered node");
+    }
+    type_ID = element_name;
+
+    if (element_ID)
+    {
+      throw RuntimeError("Attribute [ID] is not allowed in <", type_ID, ">");
+    }
   }
   else
   {
-    ID = element_name;
+    // in this case, it is mandatory to have a field "ID"
+    if (!element_ID)
+    {
+      throw RuntimeError("Attribute [ID] is mandatory in <", type_ID, ">");
+    }
+    type_ID = element_ID;
   }
 
-  const char* attr_alias = element->Attribute("name");
-  if (attr_alias)
+  // By default, the instance name is equal to ID, unless the
+  // attribute [name] is present.
+  const char* attr_name = element->Attribute("name");
+  std::string instance_name = attr_name ? attr_name : type_ID;
+
+  const TreeNodeManifest* manifest = nullptr;
+
+  auto manifest_it =  factory.manifests().find(type_ID);
+  if(manifest_it != factory.manifests().end())
   {
-    instance_name = attr_alias;
-  }
-  else
-  {
-    instance_name = ID;
+    manifest = &manifest_it->second;
   }
 
   PortsRemapping port_remap;
-
-  if (element_name == "SubTree" || element_name == "SubTreePlus")
+  for (const XMLAttribute* att = element->FirstAttribute(); att; att = att->Next())
   {
-    instance_name = element->Attribute("ID");
+    if (IsAllowedPortName(att->Name()))
+    {
+      const std::string attribute_name = att->Name();
+      port_remap[attribute_name] = att->Value();
+    }
+  }
+
+  NodeConfig config;
+  config.blackboard = blackboard;
+  config.path = prefix_path + instance_name;
+  config.uid = output_tree.getUID();
+  config.manifest = manifest;
+
+  if(type_ID == instance_name)
+  {
+    config.path += std::string("::") + std::to_string(config.uid);
+  }
+
+  auto AddCondition = [&](auto& conditions, const char* attr_name, auto ID) {
+    if (auto script = element->Attribute(attr_name))
+    {
+      conditions.insert({ID, std::string(script)});
+    }
+  };
+
+  for(int i=0; i<int(PreCond::COUNT_); i++)
+  {
+    auto pre = static_cast<PreCond>(i);
+    AddCondition(config.pre_conditions, toStr(pre).c_str(), pre);
+  }
+  for(int i=0; i<int(PostCond::COUNT_); i++)
+  {
+    auto post = static_cast<PostCond>(i);
+    AddCondition(config.post_conditions, toStr(post).c_str(), post);
+  }
+
+  //---------------------------------------------
+  TreeNode::Ptr new_node;
+
+  if (node_type == NodeType::SUBTREE)
+  {
+    config.input_ports = port_remap;
+    new_node = factory.instantiateTreeNode(instance_name,
+                                           toStr(NodeType::SUBTREE),
+                                           config);
+    auto subtree_node = dynamic_cast<SubTreeNode*>(new_node.get());
+    subtree_node->setSubtreeID(type_ID);
   }
   else
   {
-    // do this only if it NOT a Subtree
-    for (const XMLAttribute* att = element->FirstAttribute(); att; att = att->Next())
+    if(!manifest)
     {
-      const std::string attribute_name = att->Name();
-      if (ReservedPortNames.count(attribute_name) == 0)
-      {
-        port_remap[attribute_name] = att->Value();
-      }
+      throw RuntimeError("Missing manifest. It shouldn't happen. Please report this issue");
     }
-  }
-  NodeConfiguration config;
-  config.blackboard = blackboard;
-
-  //---------------------------------------------
-  TreeNode::Ptr child_node;
-
-  if (factory.builders().count(ID) != 0)
-  {
-    const auto& manifest = factory.manifests().at(ID);
 
     //Check that name in remapping can be found in the manifest
     for (const auto& remap_it : port_remap)
     {
-      if (manifest.ports.count(remap_it.first) == 0)
+      if (manifest->ports.count(remap_it.first) == 0)
       {
-        throw RuntimeError("Possible typo? In the XML, you tried to remap port "
-                           "\"",
-                           remap_it.first, "\" in node [", ID, " / ", instance_name,
-                           "], but the manifest of this node does not contain a "
-                           "port with "
-                           "this name.");
+        throw RuntimeError("Possible typo? In the XML, you tried to remap port \"",
+                           remap_it.first, "\" in node [", type_ID, " / ", instance_name,
+                           "], but the manifest of this node does not contain a port "
+                           "with this name.");
       }
     }
 
     // Initialize the ports in the BB to set the type
-    for (const auto& port_it : manifest.ports)
+    for (const auto& [port_name, port_info] : manifest->ports)
     {
-      const std::string& port_name = port_it.first;
-      const auto& port_info = port_it.second;
-
       auto remap_it = port_remap.find(port_name);
       if (remap_it == port_remap.end())
       {
         continue;
       }
-      StringView param_value = remap_it->second;
+      StringView remapped_port = remap_it->second;
 
-      if (auto param_res = TreeNode::getRemappedKey(port_name, param_value))
+      if (auto param_res = TreeNode::getRemappedKey(port_name, remapped_port))
       {
         // port_key will contain the key to find the entry in the blackboard
         const auto port_key = static_cast<std::string>(param_res.value());
@@ -574,15 +633,15 @@ TreeNode::Ptr XMLParser::Pimpl::createNodeFromXML(const XMLElement* element,
         // if the entry already exists, check that the type is the same
         if (auto prev_info = blackboard->portInfo(port_key))
         {
+          // Check consistency of types.
           bool const port_type_mismatch = (prev_info->isStronglyTyped() &&
                                            port_info.isStronglyTyped() &&
-                                           *prev_info->type() != *port_info.type());
+                                           prev_info->type() != port_info.type());
 
           // special case related to convertFromString
-          bool const string_input = ( prev_info->type() &&
-              *prev_info->type() == typeid(std::string));
+          bool const string_input = (prev_info->type() == typeid(std::string));
 
-          if (port_type_mismatch && !string_input)
+          if(port_type_mismatch && !string_input)
           {
             blackboard->debugMessage();
 
@@ -600,12 +659,12 @@ TreeNode::Ptr XMLParser::Pimpl::createNodeFromXML(const XMLElement* element,
       }
     }
 
-    // use manifest to initialize NodeConfiguration
+    // Set the port direction in config
     for (const auto& remap_it : port_remap)
     {
       const auto& port_name = remap_it.first;
-      auto port_it = manifest.ports.find(port_name);
-      if (port_it != manifest.ports.end())
+      auto port_it = manifest->ports.find(port_name);
+      if (port_it != manifest->ports.end())
       {
         auto direction = port_it->second.direction();
         if (direction != PortDirection::OUTPUT)
@@ -620,143 +679,126 @@ TreeNode::Ptr XMLParser::Pimpl::createNodeFromXML(const XMLElement* element,
     }
 
     // use default value if available for empty ports. Only inputs
-    for (const auto& port_it : manifest.ports)
+    for (const auto& port_it : manifest->ports)
     {
       const std::string& port_name = port_it.first;
       const PortInfo& port_info = port_it.second;
 
       auto direction = port_info.direction();
+
       if (direction != PortDirection::OUTPUT &&
           config.input_ports.count(port_name) == 0 &&
-          port_info.defaultValue().empty() == false)
+          !port_info.defaultValue().empty())
       {
-        config.input_ports.insert({port_name, port_info.defaultValue()});
+        try {
+          config.input_ports.insert({port_name, port_info.defaultValueString()});
+        }
+        catch(LogicError&) {}
       }
     }
 
-    child_node = factory.instantiateTreeNode(instance_name, ID, config);
-  }
-  else if (tree_roots.count(ID) != 0)
-  {
-    child_node = std::make_unique<SubtreeNode>(instance_name);
-  }
-  else
-  {
-    throw RuntimeError(ID, " is not a registered node, nor a Subtree");
+    new_node = factory.instantiateTreeNode(instance_name, type_ID, config);
   }
 
+  // add the pointer of this node to the parent
   if (node_parent)
   {
     if (auto control_parent = dynamic_cast<ControlNode*>(node_parent.get()))
     {
-      control_parent->addChild(child_node.get());
+      control_parent->addChild(new_node.get());
     }
     if (auto decorator_parent = dynamic_cast<DecoratorNode*>(node_parent.get()))
     {
-      decorator_parent->setChild(child_node.get());
+      decorator_parent->setChild(new_node.get());
     }
   }
-  return child_node;
+
+  return new_node;
 }
 
-void BT::XMLParser::Pimpl::recursivelyCreateTree(const std::string& tree_ID,
-                                                 Tree& output_tree,
-                                                 Blackboard::Ptr blackboard,
-                                                 const TreeNode::Ptr& root_parent)
+void BT::XMLParser::PImpl::recursivelyCreateSubtree(
+    const std::string& tree_ID,
+    const std::string& tree_name,
+    const std::string& prefix_path,
+    Tree& output_tree,
+    Blackboard::Ptr blackboard,
+    const TreeNode::Ptr& root_node)
 {
-  std::function<void(const TreeNode::Ptr&, const XMLElement*)> recursiveStep;
+  std::function<void(const TreeNode::Ptr&, Tree::Subtree::Ptr, std::string, const XMLElement*)>
+      recursiveStep;
 
-  recursiveStep = [&](const TreeNode::Ptr& parent, const XMLElement* element) {
+  recursiveStep = [&](TreeNode::Ptr parent_node,
+                      Tree::Subtree::Ptr subtree,
+                      std::string prefix,
+                      const XMLElement* element) 
+  {
     // create the node
-    auto node = createNodeFromXML(element, blackboard, parent);
-    output_tree.nodes.push_back(node);
+    auto node = createNodeFromXML(element, blackboard, parent_node, prefix, output_tree);
+    subtree->nodes.push_back(node);
 
-    if (node->type() == NodeType::SUBTREE)
-    {
-      if (dynamic_cast<const SubtreeNode*>(node.get()))
-      {
-        bool is_isolated = true;
-
-        for (const XMLAttribute* attr = element->FirstAttribute(); attr != nullptr;
-             attr = attr->Next())
-        {
-          if (StrEqual(attr->Name(), "__shared_blackboard"))
-          {
-            is_isolated = !convertFromString<bool>(attr->Value());
-            break;
-          }
-        }
-
-        if (!is_isolated)
-        {
-          recursivelyCreateTree(node->name(), output_tree, blackboard, node);
-        }
-        else
-        {
-          // Creating an isolated
-          auto new_bb = Blackboard::create(blackboard);
-
-          for (const XMLAttribute* attr = element->FirstAttribute(); attr != nullptr;
-               attr = attr->Next())
-          {
-            if (ReservedPortNames.count(attr->Name()) == 0)
-            {
-              new_bb->addSubtreeRemapping(attr->Name(), attr->Value());
-            }
-          }
-          output_tree.blackboard_stack.emplace_back(new_bb);
-          recursivelyCreateTree(node->name(), output_tree, new_bb, node);
-        }
-      }
-      else if (dynamic_cast<const SubtreePlusNode*>(node.get()))
-      {
-        auto new_bb = Blackboard::create(blackboard);
-        output_tree.blackboard_stack.emplace_back(new_bb);
-        std::set<StringView> mapped_keys;
-
-        for (const XMLAttribute* attr = element->FirstAttribute(); attr != nullptr;
-             attr = attr->Next())
-        {
-          const char* attr_name = attr->Name();
-          const char* attr_value = attr->Value();
-
-          if (ReservedPortNames.count(attr->Name()) != 0)
-          {
-            continue;
-          }
-          if (StrEqual(attr_name, "__autoremap"))
-          {
-            bool do_autoremap = convertFromString<bool>(attr_value);
-            new_bb->enableAutoRemapping(do_autoremap);
-            continue;
-          }
-
-          if (TreeNode::isBlackboardPointer(attr_value))
-          {
-            // do remapping
-            StringView port_name = TreeNode::stripBlackboardPointer(attr_value);
-            new_bb->addSubtreeRemapping(attr_name, port_name);
-            mapped_keys.insert(attr_name);
-          }
-          else
-          {
-            // constant string: just set that constant value into the BB
-            new_bb->set(attr_name, static_cast<std::string>(attr_value));
-            mapped_keys.insert(attr_name);
-          }
-        }
-
-        recursivelyCreateTree(node->name(), output_tree, new_bb, node);
-      }
-    }
-    else
+    // common case: iterate through all children
+    if (node->type() != NodeType::SUBTREE)
     {
       for (auto child_element = element->FirstChildElement(); child_element;
            child_element = child_element->NextSiblingElement())
       {
-        recursiveStep(node, child_element);
+        recursiveStep(node, subtree, prefix, child_element);
       }
     }
+    else   // special case: SubTreeNode
+    {
+      auto new_bb = Blackboard::create(blackboard);
+
+      std::set<StringView> mapped_keys;
+
+      for (auto attr = element->FirstAttribute(); attr != nullptr; attr = attr->Next())
+      {
+        const char* attr_name = attr->Name();
+        const char* attr_value = attr->Value();
+
+        if (StrEqual(attr_name, "_autoremap"))
+        {
+          const bool do_autoremap = convertFromString<bool>(attr_value);
+          new_bb->enableAutoRemapping(do_autoremap);
+          continue;
+        }
+        if (!IsAllowedPortName(attr->Name()))
+        {
+          continue;
+        }
+
+        if (TreeNode::isBlackboardPointer(attr_value))
+        {
+          // do remapping
+          StringView port_name = TreeNode::stripBlackboardPointer(attr_value);
+          new_bb->addSubtreeRemapping(attr_name, port_name);
+          mapped_keys.insert(attr_name);
+        }
+        else
+        {
+          // constant string: just set that constant value into the BB
+          new_bb->set(attr_name, static_cast<std::string>(attr_value));
+          mapped_keys.insert(attr_name);
+        }
+      }
+
+      std::string subtree_ID = element->Attribute("ID");
+      std::string subtree_name = subtree->instance_name;
+      if(!subtree_name.empty()) {
+        subtree_name += "/";
+      }
+      if(auto name = element->Attribute("name") ) {
+        subtree_name +=  name;
+      }
+      else {
+        subtree_name += subtree_ID + "::" + std::to_string(node->UID());
+      }
+
+      recursivelyCreateSubtree(subtree_ID,
+                               subtree_name, // name
+                               subtree_name + "/",  //prefix
+                               output_tree, new_bb, node);
+    }
   };
 
   auto it = tree_roots.find(tree_ID);
@@ -767,11 +809,19 @@ void BT::XMLParser::Pimpl::recursivelyCreateTree(const std::string& tree_ID,
 
   auto root_element = it->second->FirstChildElement();
 
-  // start recursion
-  recursiveStep(root_parent, root_element);
+  //-------- start recursion -----------
+
+  // Append a new subtree to the list
+  auto new_tree = std::make_shared<Tree::Subtree>();
+  new_tree->blackboard = blackboard;
+  new_tree->instance_name = tree_name;
+  new_tree->tree_ID = tree_ID;
+  output_tree.subtrees.push_back(new_tree);
+
+  recursiveStep(root_node, new_tree, prefix_path, root_element);
 }
 
-void XMLParser::Pimpl::getPortsRecursively(const XMLElement* element,
+void XMLParser::PImpl::getPortsRecursively(const XMLElement* element,
                                            std::vector<std::string>& output_ports)
 {
   for (const XMLAttribute* attr = element->FirstAttribute(); attr != nullptr;
@@ -779,8 +829,7 @@ void XMLParser::Pimpl::getPortsRecursively(const XMLElement* element,
   {
     const char* attr_name = attr->Name();
     const char* attr_value = attr->Value();
-    if (ReservedPortNames.count(attr_name) == 0 &&
-        TreeNode::isBlackboardPointer(attr_value))
+    if (IsAllowedPortName(attr_name) && TreeNode::isBlackboardPointer(attr_value))
     {
       auto port_name = TreeNode::stripBlackboardPointer(attr_value);
       output_ports.push_back(static_cast<std::string>(port_name));
@@ -794,98 +843,203 @@ void XMLParser::Pimpl::getPortsRecursively(const XMLElement* element,
   }
 }
 
+void addNodeModelToXML(const TreeNodeManifest& model,
+                          XMLDocument& doc,
+                          XMLElement* model_root)
+{
+  XMLElement* element = doc.NewElement(toStr(model.type).c_str());
+  element->SetAttribute("ID", model.registration_ID.c_str());
+
+  std::vector<std::string> ordered_ports;
+  PortDirection const directions[3] = {PortDirection::INPUT,
+                                       PortDirection::OUTPUT,
+                                       PortDirection::INOUT};
+  for (PortDirection direction: directions)
+  {
+    std::set<std::string> port_names;
+    for (auto& port : model.ports)
+    {
+      const auto& port_name = port.first;
+      const auto& port_info = port.second;
+      if (port_info.direction() == direction)
+      {
+        port_names.insert(port_name);
+      }
+    }
+    for (auto& port : port_names)
+    {
+      ordered_ports.push_back(port);
+    }
+  }
+
+  for (const auto& port_name : ordered_ports)
+  {
+    const auto& port_info = model.ports.at(port_name);
+
+    XMLElement* port_element = nullptr;
+    switch (port_info.direction())
+    {
+      case PortDirection::INPUT:
+        port_element = doc.NewElement("input_port");
+        break;
+      case PortDirection::OUTPUT:
+        port_element = doc.NewElement("output_port");
+        break;
+      case PortDirection::INOUT:
+        port_element = doc.NewElement("inout_port");
+        break;
+    }
+
+    port_element->SetAttribute("name", port_name.c_str());
+    if (port_info.type() != typeid(void))
+    {
+      port_element->SetAttribute("type", BT::demangle(port_info.type()).c_str());
+    }
+    if (!port_info.defaultValue().empty())
+    {
+      port_element->SetAttribute("default", port_info.defaultValueString().c_str());
+    }
+
+    if (!port_info.description().empty())
+    {
+      port_element->SetText(port_info.description().c_str());
+    }
+    element->InsertEndChild(port_element);
+  }
+
+  if (!model.description.empty())
+  {
+    auto description_element = doc.NewElement("description");
+    description_element->SetText(model.description.c_str());
+    element->InsertEndChild(description_element);
+  }
+
+  model_root->InsertEndChild(element);
+}
+
+void addTreeToXML(const Tree& tree,
+                  XMLDocument& doc,
+                  XMLElement* rootXML,
+                  bool add_metadata,
+                  bool add_builtin_models)
+{  
+  std::function<void(const TreeNode&, XMLElement*)> addNode;
+  addNode = [&](const TreeNode& node,
+                XMLElement* parent_elem)
+  {
+    XMLElement* elem = nullptr;
+
+    if (auto subtree = dynamic_cast<const SubTreeNode*>(&node))
+    {
+      elem = doc.NewElement(node.registrationName().c_str());
+      elem->SetAttribute("ID", subtree->subtreeID().c_str());
+      if(add_metadata){
+        elem->SetAttribute("_fullpath", subtree->config().path.c_str());
+      }
+    }
+    else {
+      elem = doc.NewElement(node.registrationName().c_str());
+      elem->SetAttribute("name", node.name().c_str());
+    }
+
+    if(add_metadata){
+      elem->SetAttribute("_uid", node.UID());
+    }
+
+    for(const auto& [name, value]: node.config().input_ports)
+    {
+      elem->SetAttribute(name.c_str(), value.c_str());
+    }
+    for(const auto& [name, value]: node.config().output_ports)
+    {
+      // avoid duplicates, in the case of INOUT ports
+      if(node.config().input_ports.count(name) == 0)
+      {
+        elem->SetAttribute(name.c_str(), value.c_str());
+      }
+    }
+
+    for(const auto& [pre, script]: node.config().pre_conditions)
+    {
+      elem->SetAttribute(toStr(pre).c_str(), script.c_str());
+    }
+    for(const auto& [post, script]: node.config().post_conditions)
+    {
+      elem->SetAttribute(toStr(post).c_str(), script.c_str());
+    }
+
+    parent_elem->InsertEndChild(elem);
+
+    if (auto control = dynamic_cast<const ControlNode*>(&node))
+    {
+      for (const auto& child : control->children())
+      {
+        addNode(*child, elem);
+      }
+    }
+    else if (auto decorator = dynamic_cast<const DecoratorNode*>(&node))
+    {
+      if(decorator->type() != NodeType::SUBTREE)
+      {
+        addNode(*decorator->child(), elem);
+      }
+    }
+  };
+
+  for(const auto& subtree: tree.subtrees)
+  {
+    XMLElement* subtree_elem = doc.NewElement("BehaviorTree");
+    subtree_elem->SetAttribute("ID", subtree->tree_ID.c_str());
+    subtree_elem->SetAttribute("_fullpath", subtree->instance_name.c_str());
+    rootXML->InsertEndChild(subtree_elem);
+    addNode(*subtree->nodes.front(), subtree_elem);
+  }
+
+  XMLElement* model_root = doc.NewElement("TreeNodesModel");
+  rootXML->InsertEndChild(model_root);
+
+  static const BehaviorTreeFactory temp_factory;
+
+  std::map<std::string, const TreeNodeManifest*> ordered_models;
+  for (const auto& [registration_ID, model] : tree.manifests)
+  {
+    if(add_builtin_models || !temp_factory.builtinNodes().count(registration_ID))
+    {
+      ordered_models.insert( {registration_ID, &model} );
+    }
+  }
+
+  for (const auto& [registration_ID, model] : ordered_models)
+  {
+    addNodeModelToXML(*model, doc, model_root);
+  }
+}
+
 std::string writeTreeNodesModelXML(const BehaviorTreeFactory& factory,
                                    bool include_builtin)
 {
   XMLDocument doc;
 
   XMLElement* rootXML = doc.NewElement("root");
+  rootXML->SetAttribute("BTCPP_format", "4");
   doc.InsertFirstChild(rootXML);
 
   XMLElement* model_root = doc.NewElement("TreeNodesModel");
   rootXML->InsertEndChild(model_root);
 
-  std::set<std::string> ordered_names;
+  std::map<std::string, const TreeNodeManifest*> ordered_models;
 
-  for (auto& model_it : factory.manifests())
+  for (const auto& [registration_ID, model] : factory.manifests())
   {
-    const auto& registration_ID = model_it.first;
-    if (!include_builtin && factory.builtinNodes().count(registration_ID) != 0)
+    if (include_builtin || factory.builtinNodes().count(registration_ID) == 0)
     {
-      continue;
+      ordered_models.insert( {registration_ID, &model} );
     }
-    ordered_names.insert(registration_ID);
   }
 
-  for (auto& registration_ID : ordered_names)
+  for (const auto& [registration_ID, model] : ordered_models)
   {
-    const auto& model = factory.manifests().at(registration_ID);
-
-    XMLElement* element = doc.NewElement(toStr(model.type).c_str());
-    element->SetAttribute("ID", model.registration_ID.c_str());
-
-    std::vector<std::string> ordered_ports;
-    PortDirection directions[3] = {PortDirection::INPUT, PortDirection::OUTPUT,
-                                   PortDirection::INOUT};
-    for (int d = 0; d < 3; d++)
-    {
-      std::set<std::string> port_names;
-      for (auto& port : model.ports)
-      {
-        const auto& port_name = port.first;
-        const auto& port_info = port.second;
-        if (port_info.direction() == directions[d])
-        {
-          port_names.insert(port_name);
-        }
-      }
-      for (auto& port : port_names)
-      {
-        ordered_ports.push_back(port);
-      }
-    }
-
-    for (const auto& port_name : ordered_ports)
-    {
-      const auto& port_info = model.ports.at(port_name);
-
-      XMLElement* port_element = nullptr;
-      switch (port_info.direction())
-      {
-        case PortDirection::INPUT:
-          port_element = doc.NewElement("input_port");
-          break;
-        case PortDirection::OUTPUT:
-          port_element = doc.NewElement("output_port");
-          break;
-        case PortDirection::INOUT:
-          port_element = doc.NewElement("inout_port");
-          break;
-      }
-
-      port_element->SetAttribute("name", port_name.c_str());
-      if (port_info.type())
-      {
-        port_element->SetAttribute("type", BT::demangle(port_info.type()).c_str());
-      }
-      if (!port_info.defaultValue().empty())
-      {
-        port_element->SetAttribute("default", port_info.defaultValue().c_str());
-      }
-
-      if (!port_info.description().empty())
-      {
-        port_element->SetText(port_info.description().c_str());
-      }
-      element->InsertEndChild(port_element);
-    }
-
-    if (!model.description.empty())
-    {
-      element->SetAttribute("description", model.registration_ID.c_str());
-    }
-
-    model_root->InsertEndChild(element);
+    addNodeModelToXML(*model, doc, model_root);
   }
 
   XMLPrinter printer;
@@ -901,7 +1055,8 @@ Tree buildTreeFromText(const BehaviorTreeFactory& factory, const std::string& te
   return parser.instantiateTree(blackboard);
 }
 
-Tree buildTreeFromFile(const BehaviorTreeFactory& factory, const std::string& filename,
+Tree buildTreeFromFile(const BehaviorTreeFactory& factory,
+                       const std::string& filename,
                        const Blackboard::Ptr& blackboard)
 {
   XMLParser parser(factory);
@@ -909,4 +1064,19 @@ Tree buildTreeFromFile(const BehaviorTreeFactory& factory, const std::string& fi
   return parser.instantiateTree(blackboard);
 }
 
+std::string WriteTreeToXML(const Tree &tree, bool add_metadata, bool add_builtin_models)
+{
+  XMLDocument doc;
+
+  XMLElement* rootXML = doc.NewElement("root");
+  rootXML->SetAttribute("BTCPP_format", 4);
+  doc.InsertFirstChild(rootXML);
+
+  addTreeToXML(tree, doc, rootXML, add_metadata, add_builtin_models);
+
+  XMLPrinter printer;
+  doc.Print(&printer);
+  return std::string(printer.CStr(), size_t(printer.CStrSize() - 1));
+}
+
 }   // namespace BT
